/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.control.io;

import java.util.List;
import java.util.Stack;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;
import java.net.URL;
import java.net.URLConnection;
import javax.swing.JOptionPane;

import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.XMLResourceIdentifier;
import org.apache.xerces.xni.parser.XMLInputSource;
import org.apache.xerces.xni.parser.XMLEntityResolver;

import org.xmloperator.util.io.URI;
import org.xmloperator.util.io.HttpImpl;
import org.xmloperator.util.io.FileUtils;
import org.xmloperator.framework.util.Proxy;
import org.xmloperator.framework.option.Options;
import org.xmloperator.framework.document.parse.ParseError;
import org.xmloperator.framework.document.parse.ParseErrorList;
import org.xmloperator.framework.document.parse.ToolEntityResolver;
import org.xmloperator.framework.document.doctype.DTDInfo;
import org.xmloperator.framework.document.doctype.DocumentTypeInfo;
import org.xmloperator.framework.document.FullyFledgedDocument;
import org.xmloperator.framework.view.status.MessageDisplay;

public final class ToolEntityResolverImpl implements ToolEntityResolver {
  private Options options;
  private MessageDisplay messageDisplay;
  private ParseErrorList parseErrorList = null;
  private final Stack toolInputStreamStack = new Stack();
  private ToolInputStream currentToolInputStream = null;
    // top of stack for quick access

  public ToolEntityResolverImpl(Options options,
    MessageDisplay messageDisplay) {
    this.options = options;
    this.messageDisplay = messageDisplay;
  }

  public void closeAll() throws IOException {
    while (!toolInputStreamStack.empty())
      ((InputStream)toolInputStreamStack.pop()).close();
    currentToolInputStream = null;
  }

  private void reportError(String systemId, String message) {
    if (parseErrorList != null)
      parseErrorList.add(ParseError.ERRORTYPE_ERROR, systemId, -1, -1, message);
    else
      JOptionPane.showMessageDialog(null, message, systemId,
        JOptionPane.ERROR_MESSAGE);
  }

  private XMLInputSource resolveEntity(
      String publicId, String literalSystemId,
      String baseSystemId, String expandedSystemId)
      throws XNIException, IOException {
    if (literalSystemId == null)
      throw new IllegalArgumentException();
    if (baseSystemId == null && currentToolInputStream != null)
      baseSystemId = currentToolInputStream.getSystemId();
    if (expandedSystemId == null || baseSystemId != null)
      expandedSystemId = resolveSystemId(baseSystemId, literalSystemId);
    XMLInputSource xMLInputSource =
        new XMLInputSource(publicId, literalSystemId, baseSystemId);
    if (messageDisplay != null)
      messageDisplay.displayMessage(
          "Connecting to " + truncateSystemId(expandedSystemId));
    InputStream inputStream = expandedSystemId.startsWith("http://") ?
      httpInputStream(expandedSystemId, options != null ?
      (Proxy)options.getOption(Options.PROXY) : null) :
      fileInputStream(expandedSystemId);
    if (inputStream == null) {
      reportError(expandedSystemId, "Inaccessible entity");
      xMLInputSource.setByteStream(new EmptyInputStream());
    }
    else {
      currentToolInputStream =
          new ToolInputStream(inputStream, expandedSystemId);
      toolInputStreamStack.push(currentToolInputStream);
      xMLInputSource.setByteStream(currentToolInputStream);
      if (messageDisplay != null)
        messageDisplay.displayMessage(
            "Reading " + truncateSystemId(expandedSystemId));
    }
    return xMLInputSource;
  }

  private static final String resolveSystemId(String baseSystemId,
    String systemId) {
    // systemId is supposed to be an absolute URI.
    try {
      return new URI(systemId).toString();
    }
    catch (Exception exception) {
    }
    // systemId is supposed to be a relative URI.
    if (baseSystemId != null) try {
      URI base = new URI(baseSystemId);
      return new URI(base, systemId).toString();
    }
    catch (Exception exception) {
    }
    // systemId is not recognized.
    return systemId;
  }

  private static final String truncateSystemId(String systemId) {
    final int lengthMax = 60;
    if (systemId.length() <= lengthMax)
      return systemId;
    while (systemId.length() > lengthMax) {
      String tempSystemId = systemId.substring(1);
      int barPos = tempSystemId.indexOf("/");
      if (barPos < 0)
        break;
      systemId = tempSystemId.substring(barPos);
    }
    return "..." + systemId;
  }

  private static final InputStream httpInputStream(String systemId,
    Proxy proxy) throws IOException {
    URL url = new URL(systemId);
    URLConnection connection;
    if (proxy != null && proxy.isValid()) {
      connection = new HttpImpl(url, proxy.getAddress(), proxy.getPort());
      ((HttpImpl)connection).getHeaders();
    }
    else
      connection = url.openConnection();
    return connection.getInputStream();
  }

  private static final InputStream fileInputStream(String systemId)
    throws IOException {
    String path = FileUtils.fromSystemIdToFilePathName(systemId);
    if (path == null)
      return null;
    File file = new File(path);
    return file != null && file.exists() ? new FileInputStream(file) : null;
  }

  private class EmptyInputStream extends InputStream {

    public int read() throws IOException {
      return -1;
    }
  }

  private class ToolInputStream extends InputStream {
    private InputStream inputStream;
    private String systemId;

    public ToolInputStream(InputStream inputStream, String systemId) {
      this.inputStream = inputStream;
      this.systemId = systemId;
    }

    public String getSystemId() {
      return systemId;
    }

    //////////////////////////
    // InputStream overloading
    //////////////////////////

    public int read() throws IOException {
      if (this != currentToolInputStream && currentToolInputStream != null) {
        int streamPos = toolInputStreamStack.search(this);
        if (streamPos < 0) {
          reportError(systemId, "Reopened entity");
          toolInputStreamStack.push(this);
        }
        else
          while (--streamPos > 0)
            toolInputStreamStack.pop();
        currentToolInputStream = this;
        if (messageDisplay != null)
          messageDisplay.displayMessage(
            "Reading " + truncateSystemId(systemId));
      }
      return inputStream.read();
    }

    public int available() throws IOException {
    	return inputStream.available();
    }

    public void close() throws IOException {
      inputStream.close();
    }
  }

  ///////////////////////////////////
  // XMLEntityResolver implementation
  ///////////////////////////////////

  public XMLInputSource resolveEntity(XMLResourceIdentifier resourceIdentifier)
      throws XNIException, IOException {
    return resourceIdentifier != null ? resolveEntity(
        resourceIdentifier.getPublicId(),
        resourceIdentifier.getLiteralSystemId(),
        resourceIdentifier.getBaseSystemId(),
        resourceIdentifier.getExpandedSystemId()) : null;
  }

  ////////////////////////////////////
  // ToolEntityResolver implementation
  ////////////////////////////////////

  public XMLInputSource resolveEntity(String baseURI, String systemId)
      throws XNIException, IOException {
    return resolveEntity(null, systemId, baseURI, null);
  }

  public void setParseErrorList(ParseErrorList parseErrorList) {
    this.parseErrorList = parseErrorList;
  }
}
