/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.control.operators;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.operation.OperationException;
import org.xmloperator.framework.operation.Operation;
import org.xmloperator.framework.control.operators.Element;
import org.xmloperator.framework.control.operators.LeafElement;
import org.xmloperator.framework.control.operators.TreeElement;
import org.xmloperator.framework.control.operators.Document;
import org.xmloperator.util.dom.ElementBrowser;
import org.xmloperator.util.dom.NodesComparator;
import org.xmloperator.document.util.NodeIntervalUtils;
import org.xmloperator.operation.core.NodeSetValueOperation;
import org.xmloperator.operation.core.AttrSetValueOperation;
import org.xmloperator.operation.core.NodesInsertOperation;
import org.xmloperator.operation.core.NodesRemoveOperation;
import org.xmloperator.operation.core.NodesReplaceOperation;
import org.xmloperator.operation.core.ElementUnpackOperation;
import org.xmloperator.control.operators.DocumentImpl; // Reciprocal import

public final class ElementImpl implements Element, LeafElement, TreeElement {
  private static final String LEAF_ELEMENT = "Leaf element";
  private static final String NOT_LEAF_ELEMENT = "Not leaf element";
  private static final String NO_PARENT = "No parent";
  private static final int TYPE_ELEMENT = 0;
  private static final int TYPE_LEAF_ELEMENT = 1;
  private static final int TYPE_TREE_ELEMENT = 2;
  private int elementType = TYPE_ELEMENT;
  private DocumentImpl document;
  private org.w3c.dom.Element domElement;

  public ElementImpl(DocumentImpl document, org.w3c.dom.Element domElement) {
    if (document == null || domElement == null)
      throw new IllegalArgumentException();
    this.document = document;
    this.domElement = domElement;
  }

  public org.w3c.dom.Element getDomElement() {
    return domElement;
  }

  private boolean isAttachedToRoot() {
    org.w3c.dom.Element ancestorElement = domElement;
    org.w3c.dom.Element documentElement =
      domElement.getOwnerDocument().getDocumentElement();
    while (true) {
      if (ancestorElement == documentElement)
        return true;
      ancestorElement = (org.w3c.dom.Element)ancestorElement.getParentNode();
      if (ancestorElement == null)
        return false;
    }
  }

  private Element getElement(org.w3c.dom.Element domElement) {
    return domElement != null ? document.getElement(domElement) : null;
  }

  private void insertElement(Element element,
    org.w3c.dom.Node parentNode, org.w3c.dom.Node nextSibling)
    throws UndefinedNamespaceException, IdException, OperationException {
    if (parentNode == null)
      throw new RuntimeException(NO_PARENT);
    org.w3c.dom.Element newDomElement = ((ElementImpl)element).domElement;
    if (isAttachedToRoot()) {
      Operation operation = new NodesInsertOperation(
          this.document.getNodeManager(), parentNode, nextSibling,
          newDomElement);
      document.executeAndStoreOperation(operation);
    }
    else
      parentNode.insertBefore(newDomElement, nextSibling);
  }

  private void setLeaf() {
    if (elementType == TYPE_TREE_ELEMENT)
      throw new RuntimeException(NOT_LEAF_ELEMENT);
    elementType = TYPE_LEAF_ELEMENT;
  }

  private void setTree() {
    if (elementType == TYPE_LEAF_ELEMENT)
      throw new RuntimeException(LEAF_ELEMENT);
    elementType = TYPE_TREE_ELEMENT;
  }

  /////////////////////////
  // Element implementation
  /////////////////////////

  public Document getDocument() {
    return document;
  }

  public String getName() {
    return domElement.getNodeName();
  }

  public String getAttribute(String attributeName) {
    return domElement.getAttribute(attributeName);
  }

  public void setAttribute(String attributeName, String value)
    throws UndefinedNamespaceException, IdException, OperationException {
    if (isAttachedToRoot()) {
      Operation operation = new AttrSetValueOperation(
        document.getAttributeManager(),
        domElement, attributeName, value);
      document.executeAndStoreOperation(operation);
    }
    else
      domElement.setAttribute(attributeName, value);
  }

  public boolean isEqualTo(Element element) {
    return NodesComparator.areElementsEqualWithoutChildren(
      domElement, ((ElementImpl)element).domElement)
      && hasSameChildsThan(element);
  }

  public boolean hasSameChildsThan(Element element) {
    if (elementType != TYPE_TREE_ELEMENT
      || ((ElementImpl)element).elementType != TYPE_TREE_ELEMENT)
      return true;
    Element childElement1 = this.getFirstChildElement();
    Element childElement2 = ((TreeElement)element).getFirstChildElement();
    while (childElement1 != null && childElement2 != null) {
      if (!childElement1.isEqualTo(childElement2))
        return false;
      childElement1 = childElement1.getNextSiblingElement();
      childElement2 = childElement2.getNextSiblingElement();
    }
    return childElement1 == null && childElement2 == null;
  }

  public TreeElement getParentElement() {
    org.w3c.dom.Node parentNode = domElement.getParentNode();
    return parentNode != null
      && parentNode.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE ?
      (TreeElement)document.getElement((org.w3c.dom.Element)parentNode) : null;
  }

  public Element getPreviousSiblingElement() {
    return getElement(ElementBrowser.getPreviousSiblingElement(domElement));
  }

  public Element getPreviousSiblingElement(String elementName) {
    return getElement(
      ElementBrowser.getPreviousSiblingElement(domElement, elementName));
  }

  public Element getPreviousSiblingElement(String[] elementNames) {
    return getElement(
      ElementBrowser.getPreviousSiblingElement(domElement, elementNames));
  }

  public Element getNextSiblingElement() {
    return getElement(ElementBrowser.getNextSiblingElement(domElement));
  }

  public Element getNextSiblingElement(String elementName) {
    return getElement(
      ElementBrowser.getNextSiblingElement(domElement, elementName));
  }

  public Element getNextSiblingElement(String[] elementNames) {
    return getElement(
      ElementBrowser.getNextSiblingElement(domElement, elementNames));
  }

  public void insertNewPreviousSiblingElement(Element newElement)
    throws UndefinedNamespaceException, IdException, OperationException {
    insertElement(newElement, domElement.getParentNode(), domElement);
  }

  public void insertNewNextSiblingElement(Element newElement)
    throws UndefinedNamespaceException, IdException, OperationException {
    insertElement(newElement, domElement.getParentNode(),
      domElement.getNextSibling());
  }

  public void remove()
    throws UndefinedNamespaceException, IdException, OperationException {
    if (domElement.getParentNode() == null)
      return;
    if (isAttachedToRoot()) {
      Operation operation = new NodesRemoveOperation(
          document.getNodeManager(),domElement);
      document.executeAndStoreOperation(operation);
    }
    else
      domElement.getParentNode().removeChild(domElement);
  }

  public void replaceBy(Element newElement)
    throws UndefinedNamespaceException, IdException, OperationException {
    org.w3c.dom.Node parentNode = domElement.getParentNode();
    if (parentNode == null)
      throw new RuntimeException(NO_PARENT);
    org.w3c.dom.Element newDomElement = ((ElementImpl)newElement).domElement;
    if (isAttachedToRoot()) {
      Operation operation = new NodesReplaceOperation(
          this.document.getNodeManager(), domElement, newDomElement);
      document.executeAndStoreOperation(operation);
    }
    else
      parentNode.replaceChild(newDomElement, domElement);
  }

  public LeafElement leafElement() {
    setLeaf();
    return this;
  }

  public TreeElement treeElement() {
    setTree();
    return this;
  }

  /////////////////////////////
  // LeafElement implementation
  /////////////////////////////

  public String getText() {
    setLeaf();
    org.w3c.dom.Node childNode = domElement.getFirstChild();
    if (childNode == null)
      return "";
    if (childNode.getNextSibling() == null
      && childNode.getNodeType() == org.w3c.dom.Node.TEXT_NODE)
      return childNode.getNodeValue();
    StringBuffer sb = new StringBuffer();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case org.w3c.dom.Node.TEXT_NODE:
          sb.append(childNode.getNodeValue());
          break;
        case org.w3c.dom.Node.ENTITY_REFERENCE_NODE:
          sb.append('&');
          sb.append(childNode.getNodeName());
          sb.append(';');
          break;
      }
      childNode = childNode.getNextSibling();
    }
    return sb.toString();
  }

  public void setText(String value)
    throws UndefinedNamespaceException, IdException, OperationException {
    setLeaf();
    org.w3c.dom.Node firstChildNode = domElement.getFirstChild();
    org.w3c.dom.Node lastChildNode = domElement.getLastChild();
    if (firstChildNode != null && lastChildNode == firstChildNode
      && firstChildNode.getNodeType() == org.w3c.dom.Node.TEXT_NODE)
      if (isAttachedToRoot()) {
        Operation operation = new NodeSetValueOperation(
          document.getNodeManager(), firstChildNode, value);
        document.executeAndStoreOperation(operation);
      }
      else
        firstChildNode.setNodeValue(value);
    else {
      org.w3c.dom.Node textNode =
        domElement.getOwnerDocument().createTextNode(value);
      if (isAttachedToRoot()) {
        if (firstChildNode != null) {
          Operation operation = new NodesRemoveOperation(
              this.document.getNodeManager(),
              NodeIntervalUtils.nodesFromInterval(
              firstChildNode, lastChildNode));
          this.document.executeAndStoreOperation(operation);
        }
        Operation operation = new NodesInsertOperation(
            this.document.getNodeManager(), domElement, null, textNode);
        this.document.executeAndStoreOperation(operation);
      }
      else {
        while (true) {
          org.w3c.dom.Node childNode = domElement.getFirstChild();
          if (childNode == null)
            break;
          domElement.removeChild(childNode);
        }
        domElement.appendChild(textNode);
      }
    }
  }

  /////////////////////////////
  // TreeElement implementation
  /////////////////////////////

  public int getChildElementCount() {
    setTree();
    return ElementBrowser.getChildElementCount(domElement);
  }

  public Element getFirstChildElement() {
    setTree();
    return getElement(ElementBrowser.getFirstChildElement(domElement));
  }

  public Element getFirstChildElement(String elementName) {
    setTree();
    return getElement(
      ElementBrowser.getFirstChildElement(domElement, elementName));
  }

  public Element getFirstChildElement(String[] elementNames) {
    setTree();
    return getElement(
      ElementBrowser.getFirstChildElement(domElement, elementNames));
  }

  public Element getLastChildElement() {
    setTree();
    return getElement(ElementBrowser.getLastChildElement(domElement));
  }

  public Element getLastChildElement(String elementName) {
    setTree();
    return getElement(
      ElementBrowser.getLastChildElement(domElement, elementName));
  }

  public Element getLastChildElement(String[] elementNames) {
    setTree();
    return getElement(
      ElementBrowser.getLastChildElement(domElement, elementNames));
  }

  public void appendNewChildElement(Element element)
    throws UndefinedNamespaceException, IdException, OperationException {
    setTree();
    insertElement(element, domElement, null);
  }

  public void unpack()
    throws UndefinedNamespaceException, IdException, OperationException {
    setTree();
    org.w3c.dom.Node parentNode = domElement.getParentNode();
    if (parentNode == null)
      throw new RuntimeException(NO_PARENT);
    if (isAttachedToRoot()) {
      Operation operation = new ElementUnpackOperation(
          this.document.getNodeManager(), domElement);
      document.executeAndStoreOperation(operation);
    }
    else {
      while (true) {
        org.w3c.dom.Node childNode = domElement.getFirstChild();
        if (childNode == null)
          break;
        parentNode.insertBefore(childNode, domElement);
      }
      parentNode.removeChild(domElement);
    }
  }
}
