/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.control.task;

import java.util.ResourceBundle;
import java.io.IOException;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamSource;

import org.xml.sax.SAXException;

import org.xmloperator.util.exception.ExceptionControler;
import org.xmloperator.util.io.FileUtils;
import org.xmloperator.framework.option.Constants;
import org.xmloperator.framework.option.Options;
import org.xmloperator.framework.document.transform.TransformerInfo;
import org.xmloperator.framework.view.instance.ViewAccess;
import org.xmloperator.framework.view.resource.Label;
import org.xmloperator.framework.control.task.CancelableTask;
import org.xmloperator.framework.control.task.TransformBatch;
import org.xmloperator.util.view.ExceptionDialogs;
import org.xmloperator.option.impl.TransformerInfoImpl;
import org.xmloperator.transform.XSLTTransformer;
import org.xmloperator.control.io.IOUtils;
import org.xmloperator.control.task.DocumentTransformTask;
import org.xmloperator.control.task.BatchHandler;

public final class TransformBatchTask
    implements TransformBatch, CancelableTask {

  public static final Transformer createTransformer(
      TransformerInfo transformerInfo, Options options) {
    if (transformerInfo == null)
      throw new IllegalArgumentException();
    Transformer transformer = null;
    String title = getString(Label.TRANSFORM_INFORMATION)
                 + " \"" + transformerInfo.getTransformerName() + "\"";
    String transformerClassName = transformerInfo.getTransformerClassName();
    if (transformerClassName == null || transformerClassName.length() == 0) {
      ExceptionDialogs.displayErrorMessage(title,
          getString(Label.NO_TRANSFORMER_CLASS_DEFINED));
      return null;
    }
    try {
      transformer =
          (Transformer)Class.forName(transformerClassName).newInstance();
    }
    catch (Exception exception) {
      ExceptionDialogs.displayException(exception);
      return null;
    }
    if (transformerClassName.equals(Constants.XSLT_TRANSFORMER)) {
      if (!(transformer instanceof XSLTTransformer)) {
        ExceptionDialogs.displayErrorMessage(title,
            "The transformer should be an XSLTTransformer");
        return null;
      }
      ((XSLTTransformer)transformer).setOptions(options);
      String stylesheetSystemId =
          transformerInfo.getStylesheetSystemId();
      if (stylesheetSystemId == null || stylesheetSystemId.length() == 0) {
        ExceptionDialogs.displayErrorMessage(
            title, getString(Label.NO_STYLE_SHEET_DEFINED));
        return null;
      }
      ((XSLTTransformer)transformer).setStylesheetSystemId(stylesheetSystemId);
    }
    return transformer;
  }

  private static final String getString(String key) {
    return ResourceBundle.getBundle(Label.CLASS).getString(key);
  }

  private String batchDocumentURI;
  private Options options;
  private ViewAccess viewAccess;

  public TransformBatchTask(String batchDocumentURI, Options options,
    ViewAccess viewAccess) {
    if (batchDocumentURI == null || options == null || viewAccess == null)
      throw new IllegalArgumentException();
    this.batchDocumentURI = batchDocumentURI;
    this.options = options;
    this.viewAccess = viewAccess;
  }

  private boolean executeBatch() {
    try {
      BatchHandler.process((TransformBatch)this, batchDocumentURI);
      return true;
    }
    catch (BatchHandler.BatchException exception) {
      ExceptionDialogs.displayException("Batch exception", exception);
    }
    catch (SAXException exception) {
      ExceptionDialogs.displayException("SAX exception", exception);
    }
    catch (IOException exception) {
      ExceptionDialogs.displayException(exception);
    }
    return false;
  }

  ////////////////////////////////
  // TransformBatch implementation
  ////////////////////////////////

  public boolean transform(
      String transformerName,
      String sourceDocumentURI, String destinationPathName,
      String[] nameValuePairs) {
    if (transformerName == null
        || sourceDocumentURI == null || destinationPathName == null
        || (nameValuePairs.length & 1) == 1)
      return false;
    TransformerInfo transformerInfo =
        TransformerInfoImpl.getTransformerInfo(
        options.getOptionList(Options.TRANSFORMER), transformerName);
    if (transformerInfo == null) {
      String title = getString(Label.TRANSFORM_INFORMATION)
                   + " \"" + transformerName + "\"";
      ExceptionDialogs.displayErrorMessage(
          title, getString(Label.NO_TRANSFORMER_DEFINED));
      return false;
    }
    Transformer transformer = createTransformer(transformerInfo, options);
    if (transformer == null)
      return false;
    for (int k = 0; k < nameValuePairs.length; k+= 2)
      transformer.setParameter(nameValuePairs[k], nameValuePairs[k + 1]);
    return new DocumentTransformTask(
        transformer, transformerInfo, new StreamSource(sourceDocumentURI),
        destinationPathName, options, viewAccess).run();
  }

  ////////////////////////////////
  // CancelableTask implementation
  ////////////////////////////////

  public boolean run() {
    return executeBatch();
  }

  public void cancel() {
  }
}
