/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.manage;

import java.util.List;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.document.id.IdManager;
import org.xmloperator.framework.document.schema.AttributesContent;
import org.xmloperator.framework.document.manage.AttributesContentAccess;
import org.xmloperator.framework.document.manage.AttributeChangeListener;
import org.xmloperator.framework.document.manage.AttributeManager;
import org.xmloperator.framework.document.manage.TreeViewDocumentListener;
import org.xmloperator.framework.document.manage.NodeChangeEvent;
import org.xmloperator.framework.document.manage.NodeChangeListener;
import org.xmloperator.framework.document.manage.DocumentModifiedListener;
import org.xmloperator.util.dom.NamespaceUtils;
import org.xmloperator.document.manage.NodeChangeEventImpl;

public final class AttributeManagerImpl implements AttributeManager {
  private final IdManager idManager;
  private final List nodeChangeListenerList;
  private final List documentModifiedListenerList;
  private TreeViewDocumentListener treeViewDocumentListener = null;
  private AttributesContentAccess attributesContentAccess = null;
  private AttributeChangeListener attributeChangeListener = null;

  public AttributeManagerImpl(
      IdManager idManager, List nodeChangeListenerList,
      List documentModifiedListenerList) {
    this.idManager = idManager;
    this.nodeChangeListenerList = nodeChangeListenerList;
    this.documentModifiedListenerList = documentModifiedListenerList;
  }

  public void setTreeViewDocumentListener(
      TreeViewDocumentListener treeViewDocumentListener) {
    this.treeViewDocumentListener = treeViewDocumentListener;
  }

  public void setAttributesContentAccess(
      AttributesContentAccess attributesContentAccess) {
    this.attributesContentAccess = attributesContentAccess;
  }

  public void setAttributeChangeListener(
      AttributeChangeListener attributeChangeListener) {
    this.attributeChangeListener = attributeChangeListener;
  }

  private void notifyNodeChangeEvent(NodeChangeEvent nodeChangeEvent) {
    if (this.nodeChangeListenerList != null)
      for (int k = 0; k < this.nodeChangeListenerList.size(); k++)
        ((NodeChangeListener)this.nodeChangeListenerList.get(k)).
            notifyNodeChangeEvent(nodeChangeEvent);
  }

  private void notifyDocumentModified() {
    if (this.documentModifiedListenerList == null)
      return;
    for (int k = 0; k < this.documentModifiedListenerList.size(); k++)
      ((DocumentModifiedListener)this.documentModifiedListenerList.get(k)).
          notifyDocumentModified();
  }

  //////////////////////////////////
  // AttributeManager implementation
  //////////////////////////////////

  private static final int NORMAL_ATTRIBUTE = 0;
  private static final int ID_ATTRIBUTE = 1;
  private static final int IDREF_ATTRIBUTE = 2;
  private static final int IDREFS_ATTRIBUTE = 3;

  public void setAttribute(Element element, String attrName,
    String newValue) throws IdException, UndefinedNamespaceException {
    Attr attribute = element.getAttributeNode(attrName);
    String oldValue = attribute != null ? attribute.getValue() : null;
    if (attribute == null) {
      if (newValue == null)
        return;
      String attrNamespaceURI =
        NamespaceUtils.attrQNameToNamespace(element, attrName);
      attribute = element.getOwnerDocument().
        createAttributeNS(attrNamespaceURI, attrName);
      element.setAttributeNodeNS(attribute);
    }
    int attributeType = getAttributeManagementType(element, attrName);
    switch (attributeType) {
      case ID_ATTRIBUTE:
        idManager.changeId(attribute, attribute.getValue(), newValue);
        break;
      case IDREF_ATTRIBUTE:
        idManager.removeIdRef(attribute);
        break;
      case IDREFS_ATTRIBUTE:
        idManager.removeIdRefs(attribute);
        break;
    }
    if (newValue == null)
      element.removeAttributeNode(attribute);
    else {
      attribute.setValue(newValue);
      try {
        switch (attributeType) {
          case IDREF_ATTRIBUTE:
            idManager.addIdRef(attribute);
            break;
          case IDREFS_ATTRIBUTE:
            idManager.addIdRefs(attribute);
            break;
        }
      }
      catch (IdException exception) {
        if (oldValue != null)
          attribute.setValue(oldValue);
        else
          element.removeAttributeNode(attribute);
        throw exception;
      }
    }
    if (attributeChangeListener != null)
      attributeChangeListener.notifyAttributeChange(element,
          attrName.equals("xmlns") || attrName.startsWith("xmlns:"));
    if (this.treeViewDocumentListener != null)
      this.treeViewDocumentListener.notifyAttributeSetValue(element, attrName);
    this.notifyNodeChangeEvent(
        NodeChangeEventImpl.nodeAppearanceMayHaveChangedEvent(element));
    this.notifyDocumentModified();
}

  private int getAttributeManagementType(Element element, String attrName) {
    if (this.idManager == null || this.attributesContentAccess == null)
      return NORMAL_ATTRIBUTE;
    AttributesContent attributesContent =
      attributesContentAccess.getAttributesContent(element);
    if (attributesContent == null)
      return NORMAL_ATTRIBUTE;
    String attrType = attributesContent.getType(attrName);
    if (attrType == null || attrType.indexOf(AttributesContent.TYPE_ID) != 0)
      return NORMAL_ATTRIBUTE;
    if (attrType.equals(AttributesContent.TYPE_ID))
      return ID_ATTRIBUTE;
    if (attrType.equals(AttributesContent.TYPE_IDREF))
      return IDREF_ATTRIBUTE;
    if (attrType.equals(AttributesContent.TYPE_IDREFS))
      return IDREFS_ATTRIBUTE;
    return NORMAL_ATTRIBUTE;
  }
}
