/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.relaxng;

import java.util.Iterator;
import java.util.Collection;
import java.util.Set;
import java.util.HashSet;

import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.namespace.QualifiedName;
import org.xmloperator.framework.document.relaxng.NameClass;
import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.framework.document.relaxng.GroupPattern;
import org.xmloperator.framework.document.relaxng.ElementDefinition;
import org.xmloperator.framework.document.schema.ElementContent;
import org.xmloperator.framework.document.schema.
  NotConformantToContentModelException;
import org.xmloperator.util.dom.QualifiedNameUtils;
import org.xmloperator.util.dom.NamespaceUtils;
import org.xmloperator.document.relaxng.PatternUtils;
import org.xmloperator.document.relaxng.ElementContentImpl;

public abstract class ElementDefinitionUtils {

  /////////////////
  // Pattern result
  /////////////////

  public static final NonEmptyPattern getPattern(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    return pattern != null && !pattern.isEmpty() ?
      (NonEmptyPattern)pattern : null;
  }

  /////////////////
  // Boolean result
  /////////////////

  // Return true if the element can have child nodes.
  public static final boolean canContainChilds(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    return pattern != null && !pattern.isEmpty()
      && !PatternUtils.hasOnlyAttributes((NonEmptyPattern)pattern);
  }

  // Return true if the element can have a text child node.
  public static final boolean canContainText(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    return pattern != null && !pattern.isEmpty() ?
      PatternUtils.canContainText((NonEmptyPattern)pattern) : false;
  }

  // Return true if the element should have a text child node.
  public static final boolean shouldContainText(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    return pattern != null && !pattern.isEmpty() ?
      PatternUtils.shouldContainText((NonEmptyPattern)pattern) : false;
  }

  ////////////////////////
  // ElementContent result
  ////////////////////////

  public static final ElementContent getElementContent(
    final ElementDefinition elementDefinition,
    final Object[] childElementSpecs, QualifiedName[] attributeSpecs)
    throws UndefinedNamespaceException,
    NotConformantToContentModelException {
    Pattern pattern = elementDefinition.getPattern();
    return pattern != null && !pattern.isEmpty() ?
      new ElementContentImpl((NonEmptyPattern)pattern, childElementSpecs,
      attributeSpecs) : null;
  }

  ////////////////////////////
  // Element definition result
  ////////////////////////////

  // Return amongst an element definition set the element definition
  //   which maps to an element name in the context of an element.
  public static final ElementDefinition fromElementNameToElementDefinition(
    Collection elementDefinitions, Element contextElement, String elementName)
    throws UndefinedNamespaceException {
    String prefix = NamespaceUtils.prefixFromQName(elementName);
    String localName = NamespaceUtils.localNameFromQName(elementName);
    String namespaceURI = prefix.length() != 0 ?
      NamespaceUtils.prefixToNamespace(contextElement, prefix) :
      NamespaceUtils.defaultNamespace(contextElement);
    Iterator iterator = elementDefinitions.iterator();
    while (iterator.hasNext()) {
      ElementDefinition elementDefinition = (ElementDefinition)iterator.next();
      NameClass nameClass = elementDefinition.getNameClass();
      if (NameClassUtils.isNameContained(nameClass, namespaceURI, localName))
        return elementDefinition;
    }
    return null;
  }

  ////////////////////////////////
  // Element definition set result
  ////////////////////////////////

  // Return the set of element definitions usable for element childs.
  public static final Set childElementDefinitionSet(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    if (pattern == null || pattern.isEmpty())
      return null;
    Set childElementDefinitionSet = new HashSet();
    PatternUtils.addElementDefinitions(
        childElementDefinitionSet, (NonEmptyPattern)pattern);
    return childElementDefinitionSet;
  }

  /*
  // Return the set of required element definitions usable for element childs.
  public static final Set childRequiredElementDefinitionSet(
    final ElementDefinition elementDefinition) {
    Pattern pattern = elementDefinition.getPattern();
    if (pattern == null || pattern.isEmpty())
      return null;
    Set childElementDefinitionSet = new HashSet();
    PatternUtils.addRequiredElementDefinitions(
        childElementDefinitionSet, (NonEmptyPattern)pattern);
    return childElementDefinitionSet;
  }
  */

  //////////////////////////
  // Element name set result
  //////////////////////////

  public static final Set fromElementDefinitionSetToElementNameSet(
    Element contextElement, Collection elementDefinitionSet) {
    Set qualifiedNameSet =
      fromElementDefinitionSetToQualifiedNameSet(elementDefinitionSet);
    return fromQualifiedNameSetToElementNameSet(
      contextElement, qualifiedNameSet);
  }

  private static final Set fromElementDefinitionSetToQualifiedNameSet(
    Collection elementDefinitionSet) {
    if (elementDefinitionSet == null)
      return null;
    Set qualifiedNameSet = new HashSet();
    Iterator iterator = elementDefinitionSet.iterator();
    while (iterator.hasNext()) {
      ElementDefinition elementDefinition = (ElementDefinition)iterator.next();
      NameClass nameClass = elementDefinition.getNameClass();
      NameClassUtils.appendNameClassToQualifiedNameSet(
        qualifiedNameSet, nameClass);
    }
    return qualifiedNameSet;
  }

  private static final Set fromQualifiedNameSetToElementNameSet(
    Element contextElement, Collection qualifiedNameSet) {
    if (qualifiedNameSet == null)
      return null;
    Set elementNameSet = new HashSet();
    Iterator iterator = qualifiedNameSet.iterator();
    while (iterator.hasNext()) {
      QualifiedName qualifiedName = (QualifiedName)iterator.next();
      elementNameSet.add(
          QualifiedNameUtils.getPseudoElementName(
          qualifiedName, contextElement));
    }
    return elementNameSet;
  }
}
