/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.relaxng.parse;

import java.util.Map;
import java.util.ArrayList;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.document.relaxng.impl.NonEmptyPatternImpl;
import org.xmloperator.document.relaxng.parse.RelaxNGParseException;
import org.xmloperator.document.relaxng.parse.NonEmptyPatternHandler;
import org.xmloperator.document.relaxng.parse.TextBuilder;
import org.xmloperator.document.relaxng.parse.DataBuilder;
import org.xmloperator.document.relaxng.parse.ValueBuilder;
import org.xmloperator.document.relaxng.parse.ListPatternBuilder;
import org.xmloperator.document.relaxng.parse.AttributeBuilder;
import org.xmloperator.document.relaxng.parse.RefBuilder;
import org.xmloperator.document.relaxng.parse.GroupPatternBuilder;

// nonEmptyPattern  ::= <text/>
//   | <data type="NCName" datatypeLibrary="anyURI"> param* [exceptPattern] </data>
//   | <value datatypeLibrary="anyURI" type="NCName" ns="string"> string </value>
//   | <list> pattern </list>
//   | <attribute> nameClass pattern </attribute>
//   | <ref name="NCName"/>
//   | <oneOrMore> nonEmptyPattern </oneOrMore>
//   | <choice> pattern nonEmptyPattern </choice>
//   | <group> nonEmptyPattern nonEmptyPattern </group>
//   | <interleave> nonEmptyPattern nonEmptyPattern </interleave>
public class NonEmptyPatternBuilder implements NonEmptyPatternHandler {
  private static final int STATE_EXPECT_PATTERN = 0;
  private static final int STATE_INSIDE_PATTERN = 1;
  private static final int STATE_EXPECT_END_ONEORMORE = 2;
  private static final int STATE_END = 3;

  private int state = STATE_EXPECT_PATTERN; // STATE_*
  private boolean canBeRepeated = false;
  private int oneOrMoreCount = 0;
  private NonEmptyPatternHandler nonEmptyPatternHandler = null;

  public static final boolean isStart(String elementName) {
    return elementName.equals("text")
      || elementName.equals("data")
      || elementName.equals("value")
      || elementName.equals("list")
      || elementName.equals("attribute")
      || elementName.equals("ref")
      || elementName.equals("oneOrMore")
      || GroupPatternBuilder.isStart(elementName);
  }

  public NonEmptyPatternBuilder() {
  }

  private void error() throws SAXException {
    throw new RelaxNGParseException("nonEmptyPattern", state);
  }

  ////////////////////////////////
  // ContentHandler implementation
  ////////////////////////////////

  public void startElement(String name, Attributes atts) throws SAXException {
    switch (state) {
      case STATE_EXPECT_PATTERN:
        if (name.equals("oneOrMore")) {
          canBeRepeated = true;
          oneOrMoreCount++;
          break;
        }
        else if (name.equals("text"))
          nonEmptyPatternHandler = new TextBuilder();
        else if (name.equals("data"))
          nonEmptyPatternHandler = new DataBuilder();
        else if (name.equals("value"))
          nonEmptyPatternHandler = new ValueBuilder();
        else if (name.equals("list"))
          nonEmptyPatternHandler = new ListPatternBuilder();
        else if (name.equals("attribute"))
          nonEmptyPatternHandler = new AttributeBuilder();
        else if (name.equals("ref"))
          nonEmptyPatternHandler = new RefBuilder();
        else if (GroupPatternBuilder.isStart(name))
          nonEmptyPatternHandler = new GroupPatternBuilder(new ArrayList());
        else
          error();
        state = STATE_INSIDE_PATTERN;
      case STATE_INSIDE_PATTERN:
        nonEmptyPatternHandler.startElement(name, atts);
        break;
      default:
        error();
    }
  }

  public boolean endElement() throws SAXException {
    switch (state) {
      case STATE_INSIDE_PATTERN:
        if (nonEmptyPatternHandler.endElement()) {
          if (oneOrMoreCount == 0) {
            state = STATE_END;
            return true;
          }
          state = STATE_EXPECT_END_ONEORMORE;
        }
        return false;
      case STATE_EXPECT_END_ONEORMORE:
        oneOrMoreCount--;
        if (oneOrMoreCount == 0) {
          state = STATE_END;
          return true;
        }
        return false;
      default:
        error();
        return true;
    }
  }

  public void characters(char ch[], int start, int length) {
    switch (state) {
      case STATE_INSIDE_PATTERN:
        nonEmptyPatternHandler.characters(ch, start, length);
        break;
    }
  }

  ////////////////////////////////////////
  // NonEmptyPatternHandler implementation
  ////////////////////////////////////////

  public NonEmptyPatternImpl result(Map fromNameToElementDefinition) {
    if (state != STATE_END)
      return null;
    NonEmptyPatternImpl nonEmptyPattern =
      nonEmptyPatternHandler.result(fromNameToElementDefinition);
    nonEmptyPattern.setBeRepeated(canBeRepeated);
    return nonEmptyPattern;
  }
}
