/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.operators.change;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.operation.OperationException;
import org.xmloperator.framework.control.operators.Element;
import org.xmloperator.framework.control.operators.LeafElement;
import org.xmloperator.framework.control.operators.TreeElement;
import org.xmloperator.framework.control.operators.Document;
import org.xmloperator.framework.control.operators.OperatorException;
import org.xmloperator.framework.control.operators.Operator;

public final class ChangeOperator implements Operator {

  public ChangeOperator() {
  }

  //////////////////////////
  // Operator implementation
  //////////////////////////

  public void operateElement(Element element)
    throws UndefinedNamespaceException, IdException, OperationException,
    OperatorException {
    Document document = element.getDocument();
    TreeElement changeElement = document.getRootElement();
    if (!changeElement.getName().equals("change"))
      throw new OperatorException("Root element should be <change>",
        changeElement);
    Element configurationElement = changeElement.getFirstChildElement();
    if (configurationElement == null
      || !configurationElement.getName().equals("configuration"))
      throw new OperatorException("No configuration defined", changeElement);
    if (element == null || !element.getName().equals("changesGroup")
      || element.getParentElement() != configurationElement)
      throw new OperatorException(
        "Select a main configuration changes group", configurationElement);
    executeStep(element.treeElement());
  }

  private void executeStep(TreeElement mainChangesGroupElement)
    throws UndefinedNamespaceException, IdException, OperationException,
    OperatorException {
    TreeElement configurationElement =
      mainChangesGroupElement.getParentElement();
    Document document = configurationElement.getDocument();
    // Control the version coherence
    {
      String oldVersion =
        mainChangesGroupElement.getAttribute("configVersion1");
      if (!configurationElement.getAttribute("version").equals(oldVersion))
        throw new OperatorException("Incompatible version",
          mainChangesGroupElement);
    }
    // Remove the changes groups
    {
      Element changesGroupElement =
        configurationElement.getFirstChildElement();
      while (changesGroupElement != null) {
        Element nextChangesGroupElement =
          changesGroupElement.getNextSiblingElement();
        if (changesGroupElement.getName().equals("changesGroup"))
          changesGroupElement.remove();
        changesGroupElement = nextChangesGroupElement;
      }
    }
    // Execute the changes
    {
      Element changeElement = mainChangesGroupElement.getFirstChildElement();
      while (changeElement != null) {
        String changeElementName = changeElement.getName();
        if (changeElementName.equals("moduleAdd")) {
          String moduleId = changeElement.getAttribute("id");
          String moduleName = changeElement.getAttribute("name");
          String moduleLocalisation =
            changeElement.getAttribute("localisation");
          String moduleVersion = changeElement.getAttribute("moduleVersion");
          String moduleDescription = changeElement.leafElement().getText();
          TreeElement newModuleElement =
            document.createElement("module").treeElement();
          newModuleElement.setAttribute("id", moduleId);
          newModuleElement.setAttribute("name", moduleName);
          newModuleElement.setAttribute("localisation", moduleLocalisation);
          newModuleElement.setAttribute("version", moduleVersion);
          LeafElement descriptionElement =
            document.createElement("description").leafElement();
          descriptionElement.setText(moduleDescription);
          newModuleElement.appendNewChildElement(descriptionElement);
          Element lastModuleElement =
            configurationElement.getLastChildElement("module");
          if (lastModuleElement != null)
            lastModuleElement.insertNewNextSiblingElement(newModuleElement);
          else {
            Element firstChangesGroupElement =
              configurationElement.getFirstChildElement("changesGroup");
            firstChangesGroupElement.insertNewPreviousSiblingElement(
              newModuleElement);
          }
        }
        else if (changeElementName.equals("subConfigurationAdd")) {
          String subConfigId = changeElement.getAttribute("id");
          String subConfigName = changeElement.getAttribute("name");
          String subConfigVersion =
            changeElement.getAttribute("subConfigVersion");
          Element newConfigElement = document.createElement("configuration");
          newConfigElement.setAttribute("id", subConfigId);
          newConfigElement.setAttribute("name", subConfigName);
          newConfigElement.setAttribute("version", subConfigVersion);
          Element lastSubConfigElement =
            configurationElement.getLastChildElement("configuration");
          if (lastSubConfigElement != null)
            lastSubConfigElement.insertNewNextSiblingElement(newConfigElement);
          else {
            Element firstChangesGroupElement =
              configurationElement.getFirstChildElement("changesGroup");
            firstChangesGroupElement.insertNewPreviousSiblingElement(
              newConfigElement);
          }
        }
        else if (changeElementName.equals("delete")) {
          String moduleId = changeElement.getAttribute("id");
          TreeElement moduleElement =
            document.getElement(moduleId).treeElement();
          String moduleElementName = moduleElement.getName();
          boolean isModule = moduleElementName.equals("module");
          boolean isConfiguration = moduleElementName.equals("configuration");
          if ((!isModule && !isConfiguration)
            || moduleElement.getParentElement() != configurationElement)
            throw new OperatorException(
              "Delete should refer to an immediate module or sub-configuration",
              changeElement);
          if (isConfiguration) {
            if (moduleElement.getFirstChildElement("changesGroup") != null)
              throw new OperatorException(
                "Can't delete a " + "configuration"
                + " on which " + "changes" + " are planed",
                moduleElement);
          }
          else {
            if (moduleElement.getFirstChildElement("moduleModification")
              != null)
              throw new OperatorException(
                "Can't delete a " + "module"
                + " on which " + "modifications" + " are planed",
                moduleElement);
          }
          if (document.isIdReferenced(moduleId))
            throw new OperatorException(
              "Can't delete a " + (isConfiguration ? "configuration" : "module")
              + " whose deletion is planed a second time",
              moduleElement);
          moduleElement.remove();
        }
        else if (changeElementName.equals("moduleChange")) {
          String modifId = changeElement.getAttribute("modifId");
          String moduleVersion1 = changeElement.getAttribute("moduleVersion1");
          String moduleVersion2 = changeElement.getAttribute("moduleVersion2");
          Element modifElement = document.getElement(modifId);
          TreeElement moduleElement = modifElement.getParentElement();
          if (!modifElement.getName().equals("moduleModification")
            || moduleElement == null
            || !moduleElement.getName().equals("module")
            || moduleElement.getParentElement() != configurationElement)
            throw new OperatorException(
              "Module change should refer to a module modification\n"
              + "whose module belongs to the configuration",
              changeElement);
          if (!moduleElement.getAttribute("version").equals(moduleVersion1))
            throw new OperatorException("Incompatible version", changeElement);
          if (!document.isIdReferenced(modifId))
            modifElement.remove();
          moduleElement.setAttribute("version", moduleVersion2);
        }
        else if (changeElementName.equals("subConfigurationChange")) {
          String changesGroupId = changeElement.getAttribute("changesGroupId");
          TreeElement changesGroupElement =
            document.getElement(changesGroupId).treeElement();
          TreeElement changesGroupParent =
            changesGroupElement.getParentElement();
          if (!changesGroupElement.getName().equals("changesGroup")
            || changesGroupParent == null
            || !changesGroupParent.getName().equals("configuration")
            || changesGroupParent.getParentElement() != configurationElement)
            throw new OperatorException(
              "Sub-configuration change should refer to an immediate\n"
              + "changes group of an immediate sub-configuration",
              changeElement);
          executeStep(changesGroupElement);
        }
        changeElement = changeElement.getNextSiblingElement();
      }
    }
    // Append the next changes groups
    {
      Element changesGroupElement =
        mainChangesGroupElement.getFirstChildElement();
      while (changesGroupElement != null) {
        if (changesGroupElement.getName().equals("changesGroup"))
          configurationElement.appendNewChildElement(changesGroupElement);
        changesGroupElement = changesGroupElement.getNextSiblingElement();
      }
    }
    // Update the configuration version
    {
      String newVersion =
        mainChangesGroupElement.getAttribute("configVersion2");
      configurationElement.setAttribute("version", newVersion);
    }
  }
}
