/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import java.util.List;
import java.util.ArrayList;

import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.namespace.QualifiedName;
import org.xmloperator.util.dom.NamespaceUtils;

public final class QualifiedNameImpl implements QualifiedName, Comparable {
  public static final QualifiedName EMPTY_NAME = new QualifiedNameImpl("", "");
  public static final QualifiedName ANY_NAME = new QualifiedNameImpl();

  private static final String ANY_NAMESPACE = "*";
  private String namespaceURI;
  private String localName;

  ///////////////
  // Constructors
  ///////////////

  // Representation of an AnyName
  public QualifiedNameImpl() {
    this(ANY_NAMESPACE);
  }

  // Representation of an NsName
  public QualifiedNameImpl(final String namespaceURI) {
    this(namespaceURI, "");
  }

  public QualifiedNameImpl(final String namespaceURI, final String localName) {
    if (namespaceURI == null || localName == null)
      throw new IllegalArgumentException();
    this.namespaceURI = namespaceURI;
    this.localName = localName;
  }

  public QualifiedNameImpl(final Element element)
    throws UndefinedNamespaceException {
    this(element, element.getNodeName());
  }

  public QualifiedNameImpl(
      final Element contextElement, final String elementName)
    throws UndefinedNamespaceException {
    int index = elementName.indexOf(':');
    this.localName = index >= 0 ?
      elementName.substring(index + 1) : elementName;
    String prefix = index >= 0 ? elementName.substring(0, index) : "";
    this.namespaceURI = prefix.length() == 0 ?
      NamespaceUtils.defaultNamespace(contextElement) :
      NamespaceUtils.prefixToNamespace(contextElement, prefix);
  }

  public QualifiedNameImpl(final Attr attribute)
    throws UndefinedNamespaceException {
    String attributeName = attribute.getName();
    int index = attributeName.indexOf(':');
    this.localName = index >= 0 ?
      attributeName.substring(index + 1) : attributeName;
    String prefix = index >= 0 ? attributeName.substring(0, index) : "";
    this.namespaceURI = prefix.length() != 0 ?
      NamespaceUtils.prefixToNamespace(attribute.getOwnerElement(), prefix) :
      "";
  }

  /////////////////////
  // Comparison methods
  /////////////////////

  public boolean equals(Object object) {
    return object instanceof QualifiedNameImpl ?
      localName.equals(((QualifiedNameImpl)object).localName)
      && namespaceURI.equals(((QualifiedNameImpl)object).namespaceURI) : false;
  }

  public int compareTo(Object object) {
    if (!(object instanceof QualifiedNameImpl))
      throw new IllegalArgumentException();
    int localNameCompare =
      localName.compareTo(((QualifiedNameImpl)object).localName);
    return localNameCompare != 0 ? localNameCompare :
      namespaceURI.compareTo(((QualifiedNameImpl)object).namespaceURI);
  }

  ///////////////////////////////
  // QualifiedName implementation
  ///////////////////////////////

  public String getNamespaceURI() {
    return namespaceURI;
  }

  public String getLocalName() {
    return localName;
  }

  public boolean isAnyName() {
    return namespaceURI.equals(ANY_NAMESPACE);
  }

  public boolean isNsName() {
    return localName.length() == 0;
  }

  ///////////////////////////////
  // Qualified name list creation
  ///////////////////////////////

  // Return the qualified name list of fragment children.
  // Includes elements inside entities.
  public static final QualifiedName[] qualifiedNames(final Node fragment)
    throws UndefinedNamespaceException {
    final List qualifiedNameList = new ArrayList();
    appendQualifiedNames(qualifiedNameList, fragment);
    final QualifiedName[] qualifiedNames =
      new QualifiedName[qualifiedNameList.size()];
    qualifiedNameList.toArray(qualifiedNames);
    return qualifiedNames;
  }

  // Return the qualified name list.
  // Includes elements inside entities.
  public static final QualifiedName[] qualifiedNames(final Node[] nodes)
    throws UndefinedNamespaceException {
    final List qualifiedNameList = new ArrayList();
    for (int k = 0; k < nodes.length; k++) {
      Node node = nodes[k];
      switch (node.getNodeType()) {
        case Node.ELEMENT_NODE:
          qualifiedNameList.add(new QualifiedNameImpl((Element)node));
          break;
        case Node.ENTITY_REFERENCE_NODE:
          appendQualifiedNames(qualifiedNameList, node);
          break;
      }
    }
    final QualifiedName[] qualifiedNames =
      new QualifiedName[qualifiedNameList.size()];
    qualifiedNameList.toArray(qualifiedNames);
    return qualifiedNames;
  }

  private static final void appendQualifiedNames(final List qualifiedNameList,
    final Node fragment) throws UndefinedNamespaceException {
    Node childNode = fragment.getFirstChild();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          qualifiedNameList.add(new QualifiedNameImpl((Element)childNode));
          break;
        case Node.ENTITY_REFERENCE_NODE:
          appendQualifiedNames(qualifiedNameList, childNode);
          break;
      }
      childNode = childNode.getNextSibling();
    }
  }

  // Return the qualified name list of the element attributes.
  // Includes only attributes which belong to the infoset.
  public static final QualifiedName[] attributeQualifiedNames(
    final Element element) throws UndefinedNamespaceException {
    final NamedNodeMap attributes = element.getAttributes();
    final List qualifiedNameList = new ArrayList();
    for (int k = 0; k < attributes.getLength(); k++) {
      Attr attr = (Attr)attributes.item(k);
      String attrName = attr.getName();
      if (!attrName.equals("xmlns") && !attrName.startsWith("xmlns:"))
        qualifiedNameList.add(new QualifiedNameImpl(attr));
    }
    final QualifiedName[] qualifiedNames =
      new QualifiedName[qualifiedNameList.size()];
    qualifiedNameList.toArray(qualifiedNames);
    return qualifiedNames;
  }
}
